/***************************************************************************/
/* 		This code is part of WWW grabber called pavuk		   */
/*		Copyright (c) 1997 - 2000 Stefan Ondrejicka		   */
/*		(ondrej@idata.sk)					   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#include <netdb.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#ifdef HAVE_FNMATCH
#include <fnmatch.h>
#else
#include "fnmatch.h"
#endif

#include "config.h"
#include "url.h"
#include "condition.h"
#include "tools.h"
#include "robots.h"
#include "uexit.h"
#include "re.h"
#include "dns.h"
#include "debugl.h"

typedef enum {
	CONDT_UNSUP,
	CONDT_LMAX,
	CONDT_DMAX,
	CONDT_NOFTP,
	CONDT_NOHTTP,
	CONDT_NOSSL,
	CONDT_NOGOPHER,
	CONDT_NOFTPS,
	CONDT_NOCGI,
	CONDT_ASITE,
	CONDT_DSITE,
	CONDT_ADOMAIN,
	CONDT_DDOMAIN,
	CONDT_APREFIX,
	CONDT_DPREFIX,
	CONDT_ASFX,
	CONDT_DSFX,
	CONDT_PATTERN,
	CONDT_RPATTERN,
	CONDT_SKIP_PATTERN,
	CONDT_SKIP_RPATTERN,
	CONDT_URL_PATTERN,
	CONDT_URL_RPATTERN,
	CONDT_SKIP_URL_PATTERN,
	CONDT_SKIP_URL_RPATTERN,
	CONDT_DONT_LEAVE_SITE,
	CONDT_DONT_LEAVE_DIR,
	CONDT_USER_CONDITION,
	CONDT_AIP_PATTERN,
	CONDT_DIP_PATTERN,
	CONDT_SITE_LEVEL,
	CONDT_DONT_LEAVE_SITE_ENTER_DIR,
	CONDT_LEAVE_LEVEL ,
} cond_type_t;

struct {
	cond_type_t	type;
	char 		*name;
} cond_type_info[] = {
	{CONDT_UNSUP, "unsupported"},
	{CONDT_LMAX, "-lmax"},
	{CONDT_DMAX, "-dmax"},
	{CONDT_NOFTP, "-noftp"},
	{CONDT_NOHTTP, "-nohttp"},
	{CONDT_NOSSL, "-nossl"},
	{CONDT_NOGOPHER, "-nogopher"},
	{CONDT_NOFTPS, "-noftps"},
	{CONDT_NOCGI, "-nocgi"},
	{CONDT_ASITE, "-asite"},
	{CONDT_DSITE, "-dsite"},
	{CONDT_ADOMAIN, "-adomain"},
	{CONDT_DDOMAIN, "-ddomain"},
	{CONDT_APREFIX, "-aprefix"},
	{CONDT_DPREFIX, "-dprefix"},
	{CONDT_ASFX, "-asfx"},
	{CONDT_DSFX, "-dsfx"},
	{CONDT_PATTERN, "-pattern"},
	{CONDT_RPATTERN, "-rpattern"},
	{CONDT_SKIP_PATTERN, "-skip_pattern"},
	{CONDT_SKIP_RPATTERN, "-skip_rpattern"},
	{CONDT_URL_PATTERN, "-url_pattern"},
	{CONDT_URL_RPATTERN, "-url_rpattern"},
	{CONDT_SKIP_URL_PATTERN, "-skip_url_pattern"},
	{CONDT_SKIP_URL_RPATTERN, "-skip_url_rpattern"},
	{CONDT_DONT_LEAVE_SITE, "-dont_leave_site"},
	{CONDT_DONT_LEAVE_DIR, "-dont_leave_dir"},
	{CONDT_USER_CONDITION, "-user_condition"},
	{CONDT_AIP_PATTERN, "-aip_pattern"},
	{CONDT_DIP_PATTERN, "-dip_pattern"},
	{CONDT_SITE_LEVEL, "-site_level"},
	{CONDT_DONT_LEAVE_SITE_ENTER_DIR, "-dont_leave_site_enter_dir"},
	{CONDT_LEAVE_LEVEL, "-leave_level"},
};

#ifdef DEBUG
#define COND_CHECK(type) \
{ \
	if (ret_val) \
	{ \
		if (type >= 0 && cfg.debug && cfg.debug_level && DLV12) \
		{ \
			char *urlstr = url_to_urlstr(urlp, FALSE); \
			DEBUG_LIMITS("Failed URL condition (%s) -> %s\n", \
				cond_type_info[type].name, \
				urlstr); \
			_free(urlstr);\
		} \
		return !ret_val; \
	} \
}
#else
#define COND_CHECK(type) \
{ \
        if (ret_val) \
		return !ret_val; \
}
#endif


/********************************************************/
/* kontrola ci dany server je v zozname povolenych/	*/
/* zakazanych domen					*/
/********************************************************/
static bool domain_condition(site)
char *site;
{
	char **p = priv_cfg.condition.domains;

	while(*p)
	{
		if (!(strlen(*p) > strlen(site)))
		{
			if (!strcasecmp(*p, site + strlen(site) - strlen(*p)))
			{
				return TRUE;
			}
		}
		p++;
	}
	return FALSE;
}

/********************************************************/
/* kontrola ci dokument ma dany sufix			*/
/********************************************************/
static bool cmp_sfx(urlr, sfx)
url *urlr;
char *sfx;
{
	char *pom = NULL;
	int nlen,slen;
	char *p;
	int rv;

	switch(urlr->type)
	{
		case URLT_HTTP:
#ifdef USE_SSL
		case URLT_HTTPS:
#endif
			if (urlr->p.http.searchstr)
			{
				pom = _malloc(strlen(urlr->p.http.document) +
					strlen(urlr->p.http.searchstr) + 2);
				sprintf(pom ,"%s?%s", urlr->p.http.document,
					urlr->p.http.searchstr);
				p = pom;
			}
			else
				p = urlr->p.http.document;
			break;
		case URLT_FTP:
#ifdef USE_SSL
		case URLT_FTPS:
#endif
			p = urlr->p.ftp.path;
			break;
		case URLT_GOPHER:
			p = urlr->p.gopher.selector;
			break;
		case URLT_FILE:
			p = urlr->p.file.filename;
			break;
		default: return FALSE;
	}
	
	slen = strlen(sfx);
	nlen = strlen(p);

	if (nlen < slen) return FALSE;

	rv = (!strcmp(sfx, p + nlen - slen));

	_free(pom);
	return rv;
}

/********************************************************/
/*  kontrola ci sufix URL je zo zoznamov sufixov	*/
/********************************************************/
static bool sfx_condition(urlr)
url *urlr;
{
	char **p = priv_cfg.condition.sufix;

	while (*p)
	{
		if (cmp_sfx(urlr, *p))
		{
			return TRUE;
		}
		p++;
	}

	return FALSE;
}

/********************************************************/
/* kontrola ci URL ma dany prefix			*/
/********************************************************/
static bool cmp_prefix(urlr, prefix)
url *urlr;
char *prefix;
{
	char *p;
	char *pom = NULL;
	int rv;

	switch(urlr->type)
	{
		case URLT_HTTP:
#ifdef USE_SSL
		case URLT_HTTPS:
#endif
			if (urlr->p.http.searchstr)
			{
				pom = _malloc(strlen(urlr->p.http.document) +
					strlen(urlr->p.http.searchstr) + 2);
				sprintf(pom ,"%s?%s", urlr->p.http.document,
					urlr->p.http.searchstr);
				p = pom;
			}
			else
				p = urlr->p.http.document;
			break;
		case URLT_FTP:
#ifdef USE_SSL
		case URLT_FTPS:
#endif
			p = urlr->p.ftp.path;
			break;
		case URLT_GOPHER:
			p = urlr->p.gopher.selector;
			break;
		case URLT_FILE:
			p = urlr->p.file.filename;
			break;
		default: return FALSE;
	}
	

	rv = (!strncmp(prefix, p, strlen(prefix)));
	_free(pom);
	return rv;
}


/********************************************************/
/* kontrola ci URL ma prefix zo zoznamu 		*/
/* povolenych/zakaznych					*/
/********************************************************/
static bool prefix_condition(urlr)
url *urlr;
{
	char **p = priv_cfg.condition.dir_prefix;

	while (*p)
	{
		if (cmp_prefix(urlr, *p))
		{
			return TRUE;
		}
		p++;
	}

	return FALSE;
}

/********************************************************/
/* kontola URL ci splna dany fnamtch regularny vyraz	*/
/********************************************************/
static bool cmp_pattern(urlr, pattern)
url *urlr;
char **pattern;
{
	char *p;
	char **pp;
	char *pom = NULL;

	switch(urlr->type)
	{
		case URLT_HTTP:
#ifdef USE_SSL
		case URLT_HTTPS:
#endif
			if (urlr->p.http.searchstr)
			{
				pom = tl_str_concat(NULL, urlr->p.http.document,
					"?", urlr->p.http.searchstr, NULL);
				p = pom;
			}
			else
				p = urlr->p.http.document;
			break;
		case URLT_FTP:
#ifdef USE_SSL
		case URLT_FTPS:
#endif
			p = urlr->p.ftp.path;
			break;
		case URLT_GOPHER:
			p = urlr->p.gopher.selector;
			break;
		case URLT_FILE:
			p = urlr->p.file.filename;
			break;
		default: return FALSE;
	}

	for (pp = pattern; pp && *pp ; pp++) 
	{
		if (!fnmatch(*pp, p, 0))
		{
			_free(pom);
			return FALSE;
		}
	}
	_free(pom);

	return (pattern != NULL);
}

static bool cmp_url_pattern(urlr, url_pattern)
url *urlr;
char **url_pattern;
{
	char *p = url_to_urlstr(urlr, FALSE);
	char **pp;

	for (pp = url_pattern; pp && *pp ; pp++) 
	{
		if (!fnmatch(*pp, p, 0))
		{
			_free(p);
			return FALSE;
		}
	}

	free(p);

	return (url_pattern != NULL);
}

#ifdef HAVE_REGEX
/********************************************************/
/* kontola URL ci splna dany regularny vyraz		*/
/********************************************************/
static bool cmp_rpattern(urlr, pattern)
url *urlr;
dllist *pattern;
{
	char *p;
	dllist *pp;
	char *pom = NULL;

	switch(urlr->type)
	{
		case URLT_HTTP:
#ifdef USE_SSL
		case URLT_HTTPS:
#endif
			if (urlr->p.http.searchstr)
			{
				pom = tl_str_concat(NULL, urlr->p.http.document,
					"?", urlr->p.http.searchstr, NULL);
				p = pom;
			}
			else
				p = urlr->p.http.document;
			break;
		case URLT_FTP:
#ifdef USE_SSL
		case URLT_FTPS:
#endif
			p = urlr->p.ftp.path;
			break;
		case URLT_GOPHER:
			p = urlr->p.gopher.selector;
			break;
		case URLT_FILE:
			p = urlr->p.file.filename;
			break;
		default: return FALSE;
	}

	for (pp = pattern; pp  ; pp = pp->next) 
	{
		if (re_pmatch((re_entry *)pp->data, p))
		{
			_free(pom);
			return FALSE;
		}
	}
	_free(pom);

	return (pattern != NULL);
}

static bool cmp_url_rpattern(urlr, url_pattern)
url *urlr;
dllist *url_pattern;
{
	char *p = url_to_urlstr(urlr, FALSE);
	dllist *pp;

	for (pp = url_pattern; pp ; pp = pp->next) 
	{
		if (re_pmatch((re_entry*)pp->data, p))
		{
			_free(p);
			return FALSE;
		}
	}

	free(p);

	return (url_pattern != NULL);
}

static bool cmp_ip_rpattern(urlr, ip_pattern, addr)
url *urlr;
dllist *ip_pattern;
char *addr;
{
	dllist *pp;

	for (pp = ip_pattern; pp ; pp = pp->next) 
	{
		if (re_pmatch((re_entry*)pp->data, addr))
			return FALSE;
	}
	return TRUE;
}
#endif

/********************************************************/
/* kontrola ci URL splna vsetky obmedzujuce podmienky	*/
/********************************************************/
int url_append_condition(urlp, urlnr)
url *urlp;
int urlnr;
{
	char *p;
	int ret_val;
	char *site;
#ifdef HAVE_REGEX
	bool pm1,pm2,pm3,pm4;
#endif

	ret_val = !prottable[urlp->type].supported;
	COND_CHECK(CONDT_UNSUP)

	if (!cfg.condition.ftp)
	{
		ret_val |= urlp->type == URLT_FTP;
		COND_CHECK(CONDT_NOFTP)
	}

	if (!cfg.condition.http)
	{
		ret_val |= urlp->type == URLT_HTTP;
		COND_CHECK(CONDT_NOHTTP)
	}

#ifdef USE_SSL
	if (!cfg.condition.https)
	{
		ret_val |= urlp->type == URLT_HTTPS;
		COND_CHECK(CONDT_NOSSL)
	}

	if (!cfg.condition.ftps)
	{
		ret_val |= urlp->type == URLT_FTPS;
		COND_CHECK(CONDT_NOFTPS)
	}
#endif

	if (!cfg.condition.gopher)
	{
		ret_val |=  urlp->type == URLT_GOPHER;
		COND_CHECK(CONDT_NOGOPHER)
	}

	if (cfg.condition.max_levels)
	{
		ret_val |= (urlp->level - 
			((urlp->status & URL_INLINE_OBJ) ? 1 : 0))
			 > cfg.condition.max_levels;
		COND_CHECK(CONDT_LMAX)
	}
	
	if (cfg.condition.max_documents)
	{
		ret_val |= ((!urlnr &&
			(cfg.total_cnt + 1) > cfg.condition.max_documents) || 
			(urlnr && urlnr > cfg.condition.max_documents));
		COND_CHECK(CONDT_DMAX)
	}

	if (urlp->type == URLT_GOPHER)
	{
		ret_val |= ! (urlp->p.gopher.selector[0] == '1' ||
			urlp->p.gopher.selector[0] == '0' ||
			urlp->p.gopher.selector[0] == '2' ||
			urlp->p.gopher.selector[0] == '4' ||
			urlp->p.gopher.selector[0] == '5' ||
			urlp->p.gopher.selector[0] == '6' ||
			urlp->p.gopher.selector[0] == '9' ||
			urlp->p.gopher.selector[0] == 'g' ||
			urlp->p.gopher.selector[0] == 'I');
		COND_CHECK(-1)
	}

	if (!ret_val &&
		(urlp->type == URLT_HTTP || urlp->type == URLT_HTTPS) &&
		!cfg.condition.cgi)
	{
		ret_val |= urlp->p.http.searchstr != NULL;
		COND_CHECK(CONDT_NOCGI)
	}

	site = url_get_site(urlp);
	if (!ret_val && priv_cfg.condition.sites &&
		priv_cfg.condition.sites[0] && site)
	{
		if (priv_cfg.condition.allow_site) 
		{
			ret_val |= !is_in_list(site, priv_cfg.condition.sites);
			COND_CHECK(CONDT_ASITE)
		}
		else
		{
			ret_val |= is_in_list(site, priv_cfg.condition.sites);
			COND_CHECK(CONDT_DSITE)
		}
	}
#ifdef HAVE_REGEX
	if (!ret_val && site && (priv_cfg.condition.aip ||
		priv_cfg.condition.skipip))
	{
		struct sockaddr_in addr;
		int rv;
		int is_valid = TRUE;

		_h_errno_ = 0;
		memset(&addr, '\0',  sizeof(addr));

		if ((addr.sin_addr.s_addr = inet_addr(site)) == -1)
		{
			if (dns_gethostbyname(site, &rv, (char *)&(addr.sin_addr)))
				is_valid = FALSE;
		}
		if (is_valid)
		{
			p = inet_ntoa(addr.sin_addr);
			if (priv_cfg.condition.aip)
			{
				ret_val |= cmp_ip_rpattern(urlp, priv_cfg.condition.aip, p);
				COND_CHECK(CONDT_AIP_PATTERN)
			}

			if (priv_cfg.condition.skipip)
			{
				ret_val |= !cmp_ip_rpattern(urlp, priv_cfg.condition.skipip, p);
				COND_CHECK(CONDT_DIP_PATTERN)
			}
		}
        }
#endif
	if (!ret_val && priv_cfg.condition.domains &&
		priv_cfg.condition.domains[0] && site)
	{
		if (priv_cfg.condition.allow_domain) 
		{
			ret_val |= !domain_condition(site);
			COND_CHECK(CONDT_ADOMAIN)
		}
		else
		{
			ret_val |= domain_condition(site);
			COND_CHECK(CONDT_DDOMAIN)
		}
	}

	if (!ret_val && priv_cfg.condition.sufix &&
		priv_cfg.condition.sufix[0] && urlp->type != URLT_FILE)
	{
		if (priv_cfg.condition.allow_sufix)
		{
			ret_val |= !sfx_condition(urlp);
			COND_CHECK(CONDT_ASFX)
		}
		else
		{
			ret_val |= sfx_condition(urlp);
			COND_CHECK(CONDT_DSFX)
		}
	}

	if (!ret_val && priv_cfg.condition.dir_prefix &&
		priv_cfg.condition.dir_prefix[0] && 
		urlp->type != URLT_FILE)
	{
		if (priv_cfg.condition.allow_prefix)
		{
			ret_val |= !prefix_condition(urlp);
			COND_CHECK(CONDT_APREFIX)
		}
		else
		{
			ret_val |= prefix_condition(urlp);
			COND_CHECK(CONDT_DPREFIX)
		}
	}

#ifdef HAVE_REGEX
	if (!ret_val && urlp->type != URLT_FILE)
	{
		pm1 = FALSE;
		pm2 = FALSE;
		pm3 = FALSE;
		pm4 = FALSE;

		if (priv_cfg.condition.pattern)
			pm1 = cmp_pattern(urlp, priv_cfg.condition.pattern);
		if (priv_cfg.condition.rpattern)
			pm3 = cmp_rpattern(urlp, priv_cfg.condition.rpattern);

		if (priv_cfg.condition.pattern && priv_cfg.condition.rpattern)
		{
			ret_val = pm1 && pm3;
			if (pm1)
				COND_CHECK(CONDT_PATTERN)
			else
				COND_CHECK(CONDT_RPATTERN)
		}
		else if (priv_cfg.condition.pattern)
		{
			ret_val |= pm1;
			COND_CHECK(CONDT_PATTERN)
		}
		else if (priv_cfg.condition.rpattern)
		{
			ret_val |= pm3;
			COND_CHECK(CONDT_RPATTERN)
		}
			   
		if (!ret_val)
		{
			if (priv_cfg.condition.skip_pattern)
				pm2 = !cmp_pattern(urlp, priv_cfg.condition.skip_pattern);
			if (priv_cfg.condition.rskip_pattern)
				pm4 = !cmp_rpattern(urlp, priv_cfg.condition.rskip_pattern);

			if (priv_cfg.condition.skip_pattern && priv_cfg.condition.rskip_pattern)
			{
				ret_val = pm2 && pm4;
				if (pm2)
					COND_CHECK(CONDT_SKIP_PATTERN)
				else
					COND_CHECK(CONDT_SKIP_RPATTERN)
			}
			else if (priv_cfg.condition.skip_pattern)
			{
				ret_val |= pm2;
				COND_CHECK(CONDT_SKIP_PATTERN)
			}
			else if (priv_cfg.condition.rskip_pattern)
			{
				ret_val |= pm4;
				COND_CHECK(CONDT_SKIP_RPATTERN)
			}
		}
		
	}
	if (!ret_val && urlp->type != URLT_FILE)
	{
		pm1 = FALSE;
		pm2 = FALSE;
		pm3 = FALSE;
		pm4 = FALSE;

		if (priv_cfg.condition.url_pattern)
			pm1 = cmp_url_pattern(urlp, priv_cfg.condition.url_pattern);
		if (priv_cfg.condition.rurl_pattern)
			pm3 = cmp_url_rpattern(urlp, priv_cfg.condition.rurl_pattern);

		if (priv_cfg.condition.url_pattern && priv_cfg.condition.rurl_pattern)
		{
			ret_val = pm1 && pm3;
			if (pm1)
				COND_CHECK(CONDT_URL_PATTERN)
			else
				COND_CHECK(CONDT_URL_RPATTERN)
		}
		else if (priv_cfg.condition.url_pattern)
		{
			ret_val |= pm1;
			COND_CHECK(CONDT_URL_PATTERN)
		}
		else if (priv_cfg.condition.rurl_pattern)
		{
			ret_val |= pm3;
			COND_CHECK(CONDT_URL_RPATTERN)
		}
			   
		if (!ret_val)
		{
			if (priv_cfg.condition.skip_url_pattern)
				pm2 = !cmp_url_pattern(urlp, priv_cfg.condition.skip_url_pattern);
			if (priv_cfg.condition.rskip_url_pattern)
				pm4 = !cmp_url_rpattern(urlp, priv_cfg.condition.rskip_url_pattern);

			if (priv_cfg.condition.skip_url_pattern && priv_cfg.condition.rskip_url_pattern)
			{
				ret_val = pm2 && pm4;
				if (pm1)
					COND_CHECK(CONDT_SKIP_URL_PATTERN)
				else
					COND_CHECK(CONDT_SKIP_URL_RPATTERN)
			}
			else if (priv_cfg.condition.skip_url_pattern)
			{
				ret_val |= pm2;
				COND_CHECK(CONDT_SKIP_URL_PATTERN)
			}
			else if (priv_cfg.condition.rskip_url_pattern)
			{
				ret_val |= pm4;
				COND_CHECK(CONDT_SKIP_URL_RPATTERN)
			}
		}
		
	}
#else
	if (!ret_val && priv_cfg.condition.pattern && urlp->type != URLT_FILE)
	{
		ret_val |= cmp_pattern(urlp, priv_cfg.condition.pattern);
		COND_CHECK(CONDT_PATTERN)
	}

	if (!ret_val && priv_cfg.condition.skip_pattern && urlp->type != URLT_FILE)
	{
		ret_val |= !cmp_pattern(urlp, priv_cfg.condition.skip_pattern);
		COND_CHECK(CONDT_SKIP_PATTERN)
	}

	if (!ret_val && priv_cfg.condition.url_pattern && urlp->type != URLT_FILE)
	{
		ret_val |= cmp_url_pattern(urlp, priv_cfg.condition.url_pattern);
		COND_CHECK(CONDT_URL_PATTERN)
	}

	if (!ret_val && priv_cfg.condition.skip_url_pattern && urlp->type != URLT_FILE)
	{
		ret_val |= !cmp_url_pattern(urlp, priv_cfg.condition.skip_url_pattern);
		COND_CHECK(CONDT_SKIP_URL_PATTERN)
	}
#endif

	if (!ret_val && urlp->type != URLT_FILE && cfg.condition.dont_leave_site)
	{
		url *gparent = urlp;
		bool isgp = FALSE;

		while (!isgp)
		{
#ifdef HAVE_MT
			url *ogp = gparent;
#endif
			if (gparent->status & URL_ISSTARTING)
			{
				isgp = TRUE;
				break;
			}

			LOCK_URL(ogp)
			if (gparent->parent_url)
				gparent = (url *)gparent->parent_url->data;
			else
				isgp = TRUE;
			UNLOCK_URL(ogp)
		}

		ret_val |= (urlp->type != gparent->type) || 
			(url_get_port(urlp) != url_get_port(gparent)) ||
			strcmp(url_get_site(urlp), url_get_site(gparent));
		COND_CHECK(CONDT_DONT_LEAVE_SITE)
	}

	if (!ret_val && urlp->type != URLT_FILE && cfg.condition.site_level)
	{
		url *curl = urlp;
		url *parent;
		int level = 0;
		int slevel = 0;

		LOCK_URL(urlp)
		if (urlp->parent_url)
			parent = (url *)urlp->parent_url->data;
		else
			parent = NULL;
		UNLOCK_URL(urlp)

		while (parent)
		{
			if (parent->status & URL_ISSTARTING)
				break;

			if ((curl->type != parent->type) ||
			    (url_get_port(curl) != url_get_port(parent)) ||
			    strcmp(url_get_site(curl), url_get_site(parent)))
			{
				if (!curl->moved_to || slevel)
					level++;
				slevel = 0;
			}
			else if (!curl->moved_to)
				slevel ++;
			curl = parent;
			LOCK_URL(curl)
			if (urlp->parent_url)
				parent = (url *)parent->parent_url->data;
			else
				parent = NULL;
			UNLOCK_URL(curl)
		}

		ret_val |= level > cfg.condition.site_level;
		COND_CHECK(CONDT_SITE_LEVEL)
	}

	if (!ret_val && urlp->type != URLT_FILE && cfg.condition.dont_leave_dir)
	{
		url *gparent = urlp;
		char *p1,*p2;
		int len = 0;
		bool isgp = FALSE;

		while (!isgp)
		{
#ifdef HAVE_MT
			url *ogp = gparent;
#endif
			if (gparent->status & URL_ISSTARTING)
			{
				isgp = TRUE;
				break;
			}

			LOCK_URL(ogp)
			if (gparent->parent_url)
				gparent = (url *)gparent->parent_url->data;
			else
				isgp = TRUE;
			UNLOCK_URL(ogp)
		}

		p1 = url_get_path(urlp);
		p2 = url_get_path(gparent);
		
		p = strrchr(p2, '/');
		if (p) len = p - p2;

		ret_val |= (urlp->type != gparent->type) || 
			(url_get_port(urlp) != url_get_port(gparent)) ||
			strcmp(url_get_site(urlp), url_get_site(gparent)) ||
			strncmp(p1, p2, len);
		COND_CHECK(CONDT_DONT_LEAVE_DIR)
	}

	if (!ret_val && urlp->type != URLT_FILE && cfg.condition.leave_level)
	{
		url *gparent = urlp;
		url *pomurl = urlp;
		int level = -1;
		bool isgp = FALSE;

		while (!isgp)
		{
#ifdef HAVE_MT
			url *ogp = gparent;
#endif
			if (gparent->status & URL_ISSTARTING)
			{
				isgp = TRUE;
				break;
			}

			LOCK_URL(ogp)
			if (gparent->parent_url)
				gparent = (url *)gparent->parent_url->data;
			else
				isgp = TRUE;
			UNLOCK_URL(ogp)
		}

		while (pomurl)
		{
#ifdef HAVE_MT
			url *tempurl = pomurl;
#endif
			if ((pomurl->type == gparent->type) && 
				(url_get_port(pomurl) == url_get_port(gparent)) &&
				!strcmp(url_get_site(pomurl), url_get_site(gparent)))
			{
				break;
			}

			if (!pomurl->moved_to)
				level ++;

			if ((level - ((urlp->status & URL_INLINE_OBJ) ? 1 : 0)) >=
				cfg.condition.leave_level)
			{
				ret_val = TRUE; 
				break;
			}

			if (pomurl->status & URL_ISSTARTING)
			{
				pomurl = NULL;
			}
			else
			{
				LOCK_URL(tempurl)
				if (pomurl->parent_url)
					pomurl = (url *)pomurl->parent_url->data;
				else
					pomurl = NULL;
				UNLOCK_URL(tempurl)
			}
		}
		COND_CHECK(CONDT_LEAVE_LEVEL)
	}

	if (!ret_val && urlp->type != URLT_FILE && cfg.condition.dont_leave_site_dir)
	{
		url *gparent = urlp;
		char *p1,*p2;
		int len = 0;
		bool isgp = FALSE;

		while (!isgp)
		{
			url *ogp = gparent;

			if (gparent->status & URL_ISSTARTING)
			{
				isgp = TRUE;
			}
			else
			{
				LOCK_URL(ogp)
				if (gparent->parent_url)
					gparent = (url *)gparent->parent_url->data;
				else
					isgp = TRUE;
				UNLOCK_URL(ogp)
			}

			if ((ogp->type != gparent->type) ||
			    (url_get_port(urlp) != url_get_port(gparent)) ||
			    strcmp(url_get_site(urlp), url_get_site(gparent)))
			{
				gparent = ogp;
				break;
			}
		}

                while (gparent->moved_to)
                {
                        if (gparent == urlp)
                                break;
                        gparent = gparent->moved_to;
                }

		if (!isgp)
		{
			p1 = url_get_path(urlp);
			p2 = url_get_path(gparent);
			
			p = strrchr(p2, '/');
			if (p) len = p - p2;

			ret_val |= (urlp->type != gparent->type) || 
				(url_get_port(urlp) != url_get_port(gparent)) ||
				strcmp(url_get_site(urlp), url_get_site(gparent)) ||
				strncmp(p1, p2, len);
			COND_CHECK(CONDT_DONT_LEAVE_SITE_ENTER_DIR)
		}
	}

	if (!ret_val && priv_cfg.condition.uexit)
	{
		ret_val = !uexit_condition(urlp, NULL, 0L);
		COND_CHECK(CONDT_USER_CONDITION)
	}

	return !ret_val;
}

